#############################################################################
##
#W  fpcomsmg.gi           GAP library         Isabel Araujo and Andrew Solomon
##
#H  @(#)$Id: fpcomsmg.gi,v 1.2 1999/07/21 18:11:00 gap Exp $
##
#Y  Copyright (C)  1997,  Lehrstuhl D fuer Mathematik,  RWTH Aachen,  Germany
#Y  (C) 1998 School Math and Comp. Sci., University of St.  Andrews, Scotland
##
##  This file contains code for commutative semigroups
##

Revision.fpcomsmg_gi :=
    "@(#)$Id: fpcomsmg.gi,v 1.2 1999/07/21 18:11:00 gap Exp $";


###############################################################
##
##	From the theory we know that the commutative fp smg S is 
##	finite iff any crws has, for all generators x,
##	a relation which lhs is x^n, for some n
##	
##	The exponents of this relations give us a bound on the
##	size of the commutative semigroup, since if x^n is a lhs
##	then every word with an occurence of x^k, k>=n, is
##	necessarily equal to a smaller one
##	We build a set T of vectors: all vectors with i-th entry
##	less than the minimum exponent to which generator i
##	appears in relations with lhs of the type x^n
##	The size of T is the refered bound on the size of S
##
##	But other relations from r may give rise to vectors
##	of T representing the same word.
##	ie, S may have strictly less elements than T
##
##	In fact, we have to remove the size of the union of the 
##	ideals generated by lhs of those other relations, because 
##	they correpond to words which are not reduced.
##	
##	If the union of such ideals of vectors ir I then
##	|S|=|T|-|I|
##	Then we just have to calculate the size of T
##	and the size of I. The latter is done by a combinatorial
## 	argument (inclusion/exclusion)
##
##	In this way we calculate the size of the semigroup S,
##	given a crws for it, without enumerating S
##

#############################################################################
##
#M  EpimorphismAbelianization(<S>)
##
##  returns an epimorphism from an fp semigroup to its abelianization.
##
InstallMethod(EpimorphismAbelianization,
"<S>", true,
[IsFpSemigroup],0,
function(s)

	local f,		# free semigroup
				i,j,	# loop variables 
				r,		# relations 
				x,		# generators of semigroup
				epifun, # the function which computes the epimorphism
				cs;		# the abelianized semigroup

	#add relations to make it commutative
	#relations added are: x*y=y*x for x,y generators of
	#the free underlying semigroup
	f:=FreeSemigroupOfFpSemigroup(s);
	x:=GeneratorsOfSemigroup(f);
	r:=ShallowCopy(RelationsOfFpSemigroup(s));
	
	for i in [1..Length(x)] do
		for j in [1..i-1] do
			Add(r,[x[j]*x[i],x[i]*x[j]]);
		od;
	od;

	cs:=f/r;
	SetIsCommutative(cs,true);

	epifun :=  g->
	VectorToElementOfCommutativeFpSemigroup(cs, ElementOfFpSemigroupAsVector(g));

	return MagmaHomomorphismByFunctionNC(s, cs, epifun);

end);

#############################################################################
##
#M  Abelianization(<S>)
##
InstallMethod(Abelianization,
"<S>", true,
[IsFpSemigroup],0,
function(s)

	#if gap already knows that s is a commutative semigroup,
	#it returns s itself 
	if HasIsCommutative(s) and IsCommutative(s) then
		return s;
	fi;

	return Range(EpimorphismAbelianization(s));
end);


#############################################################################
##
#F  AssocWordToVector (<w>)
##  
##	given a word in the Free Semigroup, transforms it into a vector,
##	where entry i corresponds to the number
##	of times that generator i appears in the word
##
InstallGlobalFunction(AssocWordToVector,
function(w)

	local i,			# loop variable
				k,			# generator of syllable
				m,			# number of generators
				n,			# number of syllables
				vector;	# resulting vector

	#the vector will have length equal to no of generators of smg
	m:=Length(MagmaGeneratorsOfFamily(FamilyObj(w)));
	n:=NumberSyllables(w);
	vector:=[];

	#we start by having zeros in all entries
	for i in [1..m] do Add(vector,0);od;

	#run through the syllables of the word
	#and add the exponent to the entry of the vector
	#correponding to the syllable generator
	for i in [1..n] do
		k:=GeneratorSyllable(w,i);
		vector[k]:=vector[k]+ExponentSyllable(w,i);
	od;
	
	return vector;	

end);
	
#############################################################################
##
#F  ElementOfFpSemigroupAsVector (<e>)
##  
##	
InstallGlobalFunction(ElementOfFpSemigroupAsVector,
function(e)
	local w;	# underlying word

	w:=UnderlyingElement(e);

	return AssocWordToVector(w);
end);

#############################################################################
##
#F  VectorToElementOfCommutativeFpSemigroup(<S>,<v>)
##  
##
InstallGlobalFunction(VectorToElementOfCommutativeFpSemigroup,
function(s,v)

local 
        allentrieszero,	# true iff all entries of vector are zero
	f,	# the family to which the smg elms belong
	fam,	# the family to which the free generators of the smg belong
	i,	# loop variable
	t,	# type of generators of semigroup
        vl,     # the external representation of the desired element
	x;	# generators of semigroup

	#the following checks if a vector is the zero vector
	allentrieszero:=function(v)
		local aux,i;

		aux:=true;
		i:=1;
		for i in [1..Length(v)] do
			if v[i]<>0 then
				return false;
			fi;
		od;
		return aux;
	end;

	x:=FreeGeneratorsOfFpSemigroup(s);
	
	if Length(x)<>Length(v) then
		Error("Wrong vector length");
	fi;

	#if the semigroup is not a monoid, or does not know
	#that is a monoid, then zero vectors are not aloud
	if (not IsMonoid(s)) then
		if allentrieszero(v) then
			Error("Zero vector not aloud (semigroup does not have identity)");
		fi;
	fi;

  # we now build the external representation of the free object
  # underlying the element of the fp semigroup
  vl := List([1..2*Length(v)],function(i)
     if IsOddInt(i) then return (i+1)/2; else return v[i/2]; fi;
     end);
  fam :=  FamilyObj(x[1]);
  f:=FamilyObj(GeneratorsOfSemigroup(s)[1]);

  return ElementOfFpSemigroup(f,(ObjByExtRep(fam,vl)));

end);

#############################################################################
##
#F  VectorOfSupOfEntriesOfElementsOfCommutativeFpSemigroup(<s>)
##  
##	Returns a vector which entries are sup of entries of reduced words,
##	ie, sups of entries of elements of the semigroup.
##
InstallGlobalFunction(
VectorOfSupOfEntriesOfElementsOfCommutativeFpSemigroup,
function(s)

	local i,			# loop variable
				rws,		# KnuthBendix rws
				lhs,		# lhs of the relation being looked at
				m,			# number of generators of the semigroup
				r,			# Rules of the rws
				vector;	# the vector of sup
	
	if not(IsCommutative(s)) then
		Error("Semigroup should be commutative");
	fi;

	m:=Length(FreeGeneratorsOfFpSemigroup(s));

	#get a reduced confluent rws for s
	rws:=KnuthBendixRewritingSystem(s);
	MakeConfluent(rws);
	#just get the relations
	r:=Rules(rws);

	vector:=[];
	for i in [1..Length(r)] do
		lhs:=r[i][1];
		if NumberSyllables(lhs)=1 then
			vector[GeneratorSyllable(lhs,1)]:=ExponentSyllable(lhs,1);
		fi;
	od;

	#if there are undound entries in the vector, then it means
	#that some entries are not bound (ie, smg is infinite). 
	#In this case return fail
	for i in [1..m] do
		if not(IsBound(vector[i])) then
			return fail;
		fi;
	od;

	return vector;
	
end);
			

#############################################################################
##
#M  IsFinite( <S> ) . . . . . . . . . . . . . for a fp commutative semigroup
##
InstallMethod( IsFinite,
    "method for a fp commutative semigroup",
    true,
    [ IsFpSemigroup and IsCommutative ], 0,
		function(S)
    return VectorOfSupOfEntriesOfElementsOfCommutativeFpSemigroup(S)<>fail;
		end);


############################################################################
##
#F  SizeOfFpCommutativeSemigroup(<S>)
##  
##	for a fp commutative semigroup S 
##
##	the global function to calculate the size of an abelian fp semigroup
##	when we have the proper category this function should be a function
##	of a commutative semigroup
##
##	This function takes a fp commutative semigroup 
##
InstallGlobalFunction(SizeOfFpCommutativeSemigroup,
function(s)

	local auxsize,							# size of union of ideals 					 
				auxsum,								# sum of sizes of intersection of ideals
				generators_of_ideals, # function of the rules, returning gens of ideals
				i,j,U,								# loop variables
				idealgenerators,			# set of ideal generators 
				idgens,								# subset of ideal generators
				i_subsets_of_n,				# list of subsets of {1..n} with i elements 
				kbrws,								# KnuthBendix Rws 
				r,										# set of rules
				rhovector,						# vector of sup of entries
				size_of_intersection_of_ideals,	
				t;										# upper bound for size of semigroup
				

	#given a set of rules (of a reduced confluent kbrws)
	#this function returns the vectors which are generators of
	#ideals of non-reduced words (with more than one non-zero component
	#(ie, vectors correponding to left hand sides of relations of
	#Rules(rs), which are neither trivial nor involve only one generator)
	generators_of_ideals:=function(r)

		local listgen,u,v;

		listgen:=[];
		for u in r do
			v:=[AssocWordToVector(u[1]),AssocWordToVector(u[2])];
			if v[1]<>v[2] then
				if NumberSyllables(u[1])>1 then
					Add(listgen,AssocWordToVector(u[1]));
				fi;
			fi;
		od;

		return Set(listgen);

	end;

	size_of_intersection_of_ideals:=function(rhovector,idealgenerators)

		local auxcount,auxmax,i,k,vectorofmax;

		#if a vector is in the intersection then all its entries
		#have to be greater then or equal to the correponding
		#entries of all the ideal generators. The intersection will be 
		#exactly the set of all vectors with entry i 
		#max of entries i of generators of ideals<=i<rhovector(i)

		#we first compute the vector which entries are
		#the maximum of correponding entries of all
		#generators of the considered ideals
		vectorofmax:=[];
		for i in [1..Length(rhovector)] do
			auxmax:=0;
			for k in [1..Length(idealgenerators)] do
				auxmax:=Maximum(auxmax,idealgenerators[k][i]);
			od;
			vectorofmax[i]:=auxmax;
		od;

		#then count how many elements are in the intersection
		#(elements with entry i satisfied the above stated)
		auxcount:=1;
		for i in [1..Length(rhovector)] do
			auxcount:=auxcount*(rhovector[i]-vectorofmax[i]);
		od;

		return auxcount;

	end;

	#first check if the semigroup is finite or infinite
	if not(IsFinite(s)) then
		return infinity;
	fi;

	#First we need the set of relations of a reduced confluent
	#kb rws for the semigroup <s>

	kbrws:=KnuthBendixRewritingSystem(s);
	MakeConfluent(kbrws);
	r:=Rules(kbrws);

	#here we know that the semigroup is finite
	#we compute the vector of sup of entries, call it 'rhovector'
	#(ie, any word in the semigroup can be represented by
	#an element with i-th entry is less than i'th entry of
	#the 'rhovector')
	rhovector:=VectorOfSupOfEntriesOfElementsOfCommutativeFpSemigroup(s);

	#from the rhovector we consider the set T which is  a set of vectors: 
	#all vectors with i-th entry less than the i entry of the rhovector. 
	#The size of T is a upper bound to the size of S	
	#We compute the size, t, of T:
	t:=1;
	for i in rhovector do
		t:=t*i;
	od;
	#now we have all elements, plus the empty word which is not allowed
	t:=t-1;
	
	#We know calculate the size of the union of all ideals of non-reduced 
	#words (these are principal ideals, generated by the vectors 
	#corresponding to the lhs of relations of the rewriting system for s.
	#We do this by inclusion/exclusion principle

	auxsize:=0;
	idealgenerators:=generators_of_ideals(r);
		
	#the following cycle is the inclusion/exclusion principle
	#(we have as many ideals as ideal generators)
	for i in [1..Length(idealgenerators)] do

		#add sum of sizes of all intersections of i ideals
		auxsum:=0;
		i_subsets_of_n:=
			Combinations(Set([1..Length(idealgenerators)]),i);
		for U in i_subsets_of_n do
			idgens:=[];
			for j in U do
				Add(idgens,idealgenerators[j]);
			od;
			auxsum:=auxsum+size_of_intersection_of_ideals(rhovector,idgens);
		od;
		auxsize:=auxsize+(-1)^(i+1)*auxsum;
	od;
	#auxsize is now the size of the union of the considered ideals.	

	#the size of S is just t-auxsize
	return t-auxsize;
	
end);

#############################################################################
##
#M  Size( <S> ) . . . . . . . . . . . . . for a fp commutative semigroup
##
InstallMethod( Size,
    "method for a fp commutative semigroup",
    true,
    [ IsFpSemigroup and IsCommutative ], 0,
		function(s)
			local sizeofs;
			sizeofs:=SizeOfFpCommutativeSemigroup(s);
			return sizeofs;
		end);

#############################################################################
##
#F  ReducedSetOfGeneratorsOfPrincipalIdealOfSemigroup( <S> , <e> )
##
InstallGlobalFunction(ReducedSetOfGeneratorsOfPrincipalIdealOfSemigroup,
function(s,v)
local i,                          # loop variable
      rws,                        # the Knuth Bendix rws
      U,                          # generators of free semigroup
      vt,                         # reduced rep of v
      V,                          # gen for the ideal
      forward,                    # apply rules to els of V
      Minimize;                   # minimize wrt division of vectors

  ##########################################################
  # for each element of V and each rule of the rws
  # multiply  the element of V by something in such a way
  # that is divisible (as a vector) by the lhs of the rule.
  # Then apply rule (in the normal way).
  forward:=function(V,rws)

    local min_missing,  # what is missing so that 2nd divides first arg
          missing,      # stores result of min_missing
          set,          # resulting set
          u,            # element of V
          rhsvector,    # rhs of rule in the form of a vector
          lhsvector,    # lhs of rule in the form of a vector
          uvector,      # u in the form of a vector
          i,            # loop variable
          w;

  	###############################################
    #this funtion computes the minimum vector
    #that when added to w will produce a vector
    #which rhs divides (ie, a vector will all entries
    #greater than or equal the ones in rhs)
    min_missing:=function(wvector,rhsvector)

      local i,                # loop variable
            m,                # length of vectors
            whatismissing;    # the result

 
      whatismissing:=[];
      m:=Length(wvector);
      for i in [1..m] do
        if rhsvector[i]>wvector[i] then
          whatismissing[i]:=rhsvector[i]-wvector[i];
        else
          whatismissing[i]:=0;
        fi;
      od;

      return whatismissing;

    end;

    set:=[];
    for u in V do

      for i in [1..Length(Rules(rws))] do
        uvector:=AssocWordToVector(u);
        rhsvector:=AssocWordToVector(Rules(rws)[i][2]);
        lhsvector:=AssocWordToVector(Rules(rws)[i][1]);
        missing:=min_missing(uvector,lhsvector);
        w:=List([1..Length(uvector)],j->
               ( (uvector[j]+missing[j]) - lhsvector[j] )+rhsvector[j]);
        w:=ReducedForm(rws,UnderlyingElement(
                VectorToElementOfCommutativeFpSemigroup(s,w)));
        Add(set,w);
      od;

    od;

    return set;

  end;

  ##########################################################
  # given a set of words returns a subset of them
  # with the minimal ones wrt the division order of the
  # corresponding vector
  Minimize:=function(gens)

    local genset,               # set of gens (ie gets rid of repetitions)
          isdivisibleby,        # true or false
          j,                    # loop variable
          nonredundant;         # the result

    ###########################################
    #this function returns true if the vector u
    #is divisible by v, ie, all entries of u are
    #greater then or equal to the entries of v
    isdivisibleby:=function(u,v)

      local uvector,vvector;    # vectors representing the elements

      # transforms the words to vectors
      uvector:=AssocWordToVector(u);
      vvector:=AssocWordToVector(v);

      # checks if all entries of u are greater than or equal to the ones of v
      return ForAll([1..Length(uvector)],i->vvector[i]<=uvector[i]);

    end;

    # get the set of gens, ie, get rid of repetitions
    genset:=Set(gens);

    # we start with an empty set and keep adding the gens that
    # are not divisible by anything else
    nonredundant:=[];

    # the following tests if the generator is divisible by
    # anything else
    for j in [1..Length(genset)] do

      if not( ForAny([1..Length(genset)],i->i<>j
          and isdivisibleby(genset[j],genset[i]))) then
        # and add it to the set
        Add(nonredundant,genset[j]);
      fi;

    od;
    return nonredundant;
  end;


  # now the actual function

  # get the Knuth Bendix rws of the semigroup and  make it confluent
  # get the reduced form of the word underlying v
  # and free generators for the semigroup
  rws:=KnuthBendixRewritingSystem(s);
  MakeConfluent(rws);
  vt:=ReduceWordUsingRewritingSystem(rws,UnderlyingElement(v));
  U:=FreeGeneratorsOfFpSemigroup(s);

  # initialization
  V:=[]; V[1]:=[]; V[2]:=[vt];
  i:=2;

  # now the cycle
  # it keeps going provided there were in V[i] new elements
  # with respect to V[i-1]
  # Next step is building V[i+1]
  while (V[i]<>V[i-1]) do

    i:=i+1;

    #V[i] consists of the previous generators (ie, V[i-1]) ...
    V[i]:=ShallowCopy(V[i-1]);

    # ... union with the set obtained by multiplying
    # each generator of V[i-1] (for each rule) by a word such that
    # generator*word is divisible by the lhs of the rule, and
    # then reduce the obtained word.
    Append(V[i],forward(V[i-1],rws));

    #We finally get rid of all non-minimal elements,
    #with respect to the division ordering on the vectors.
    V[i]:=Minimize(V[i]);

  od;
  return V[i];

end);

#############################################################################
##
#M  IsGreensLessThanOrEqual( <C1>, <C2> )
##	
##	for any greens classes (recall that L=R=J=H=D in a commutative semigroup
##
##	Let <e1> be a reduced word representing the representative of <C1>.
##	Let <e2> be a reduced word representing the representative of <C2>.
##	<C1> is less then or equal to  <C2> iff <e1> is in the ideal
##	generated by <e2> (ideal in the free semigroup), ie, iff <e1> is
##	in the ideal generated by the set 
##	<J> = ReducedSetOfGeneratorsOfPrincipalIdealOfSemigroup( <S>, <e2>)
##	So, <C1> is less then or equal to  <C2> iff there is an element z in J
##	such that <z> divides <e1> (in the vector division sense) 
##
##	The reason why the rank of this method is increased is to be sure
##	that method selection always chooses this one, for otherwise
##	in the case of R,L and J classes of commutative semigroups, the normal
##	methods would be choosen.
##
InstallMethod( IsGreensLessThanOrEqual, 
	"for greens classes of a finitely presented  commutative semigroup", IsIdenticalObj,
  [IsGreensClass, IsGreensClass], 5,
function( c1, c2 )
	local	e1, e2, s, rws, set2, ve1, z, vz;

	# get the parent semigroup and its kbrws
	s := Parent( c1 );

	# this method only works for commutaive semigroups
	if HasIsCommutative( s ) then
		if not( IsCommutative( s )) then
			TryNextMethod();
		fi;
	else
		TryNextMethod();
	fi;

	rws := KnuthBendixRewritingSystem( s );
	MakeConfluent( rws );

	# get reduced words representing the representatives of each of the classes 
	e1 := ReducedForm( rws, UnderlyingElement(Representative( c1 )));
	e2 := Representative( c2 );		

	# the vector form of v1
	ve1 := AssocWordToVector( e1 );

	# the ReducedSetOfGeneratorsOfPrincipalIdealOfSemigroup of e2
	set2 := ReducedSetOfGeneratorsOfPrincipalIdealOfSemigroup( s, e2); 

	# now browse throught set2 to look for an element that divided e1 
	for z in set2 do
		vz := AssocWordToVector( z );
		if ForAll( [1.. Length(vz)], i -> vz[i]<=ve1[i] ) then
			# if vz divides ve1 then ve1 is in the ideal generated by set2
			# and hence C1 is greens less than C2 
			return true;
		fi;
	od;

	# otherwise C1 is not less than C2
	return false;		

end);


#############################################################################
##
#E  commsmg.gi  . . . . . . . . . . . . . . . . . . . . . . . . . . ends here
